#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>

#include "spi.h"
#include "ST7735R_TFT.h"

#define START_X 25
#define START_Y 40
#define STEP_X 6
#define STEP_Y 10

uint8_t width;
uint8_t height;
uint8_t pointer_x;
uint8_t pointer_y;
uint8_t wins;
uint8_t losses;
int8_t board[8][15];
uint8_t marks[8][15];
uint8_t bombs;
uint8_t squares_left;
uint8_t start_x;
uint8_t game_on;



void draw_board()
{
    uint8_t i, j, x = start_x, y = START_Y;
    char stats[12];
    char c[2];

    for(i = 0; i < height; i++){
        for(j = 0; j < width; j++){

            if(marks[i][j] == 0)
                sprintf(c, " ");
            else if(marks[i][j] == 2)
                sprintf(c, "F");
            else{
                if(board[i][j] == -1)
                    sprintf(c, "X");
                else if(board[i][j] == 0)
                    sprintf(c, " ");
                else
                    sprintf(c, "%hhu", board[i][j]);
            }

            if(i == pointer_y && j == pointer_x)
                ST7735R_DrawText(x, y, c, 255, 255, 255, 255, 0, 0);
            else{
                if(marks[i][j] == 1)
                    ST7735R_DrawText(x, y, c, 255, 255, 255, 100, 100, 100);
                else
                    ST7735R_DrawText(x, y, c, 255, 255, 255, 0, 0, 0);    
            }
        
            x += STEP_X;
        }
        x = start_x;
        y += STEP_Y;
    }

    sprintf(stats, "W: %d | L: %d", wins, losses);
    ST7735R_DrawText(1, ST7735R_HEIGHT - STEP_Y, stats, 0, 0, 0, 255, 255, 255);
}

uint8_t is_valid(int8_t i, int8_t j)
{
    if(i < 0 || i >= height)
        return 0;
    if(j < 0 || j >= width)
        return 0;

    return 1;
}

void fill(int8_t lin, int8_t col)
{
    if(!is_valid(lin, col))
        return;

    if(board[lin][col] == 0 && marks[lin][col] == 0){
        squares_left --;
        marks[lin][col] = 1;
        fill(lin + 1, col);
        fill(lin - 1, col);
        fill(lin, col + 1);
        fill(lin, col - 1);
    }
}

void plant_bomb(int8_t i, int8_t j)
{
    if(is_valid(i + 1, j) && board[i + 1][j] >= 0)
        board[i + 1][j] ++;

    if(is_valid(i - 1, j) && board[i - 1][j] >= 0)
        board[i - 1][j] ++;

    if(is_valid(i, j + 1) && board[i][j + 1] >= 0)
        board[i][j + 1] ++;

    if(is_valid(i, j - 1) && board[i][j - 1] >= 0)
        board[i][j - 1] ++;

    if(is_valid(i + 1, j + 1) && board[i + 1][j + 1] >= 0)
        board[i + 1][j + 1] ++;

    if(is_valid(i - 1, j - 1) && board[i - 1][j - 1] >= 0)
        board[i - 1][j - 1] ++;

    if(is_valid(i + 1, j - 1) && board[i + 1][j - 1] >= 0)
        board[i + 1][j - 1] ++;

    if(is_valid(i - 1, j + 1) && board[i - 1][j + 1] >= 0)
        board[i - 1][j + 1] ++;
}

void compute_board()
{
    int8_t i, lin, col, j;

    for(i = 0; i < bombs; i++){
        lin = (uint8_t) rand() % height;
        col = (uint8_t) rand() % width;

        if( board[lin][col] == 0)
            board[lin][col] = -1;
        else
            i--;
    }

    for(i = 0; i < height; i++)
        for(j = 0; j < width; j++)
            if(board[i][j] == - 1)
                plant_bomb(i, j);
}

void game_init()
{
    uint8_t lvl = 0, i, j;
    game_on = 0;
    pointer_x = 0;
    pointer_y = 0;

    ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 255, 255, 255);
    ST7735R_DrawText(START_X, START_Y + 1 * STEP_Y, "EASY", 0, 0, 0, 255, 0, 0);
    ST7735R_DrawText(START_X, START_Y + 2 * STEP_Y, "MEDIUM", 0, 0, 0, 255, 255, 255);
    ST7735R_DrawText(START_X, START_Y + 3 * STEP_Y, "HARD", 0, 0, 0, 255, 255, 255);

    for(i = 0; i < 8; i++)
        for(j = 0; j < 15; j++){
            board[i][j] = 0;
            marks[i][j] = 0;
        }

    while(1)
    {
        if ((PINA & (1 << PA7)) != 0)
        {
            if(lvl == 0)
                lvl = 3;

            lvl -= 1;
        }
        else if ((PINA & (1 << PA4)) != 0)
        {
            lvl = (lvl + 1) % 3;
        }
        else if ((PINA & (1 << PA6)) != 0)
        {
            break;
        }

        if(lvl == 0)
        {
            ST7735R_DrawText(START_X, START_Y + 1 * STEP_Y, "EASY", 0, 0, 0, 255, 0, 0);
            ST7735R_DrawText(START_X, START_Y + 2 * STEP_Y, "MEDIUM", 0, 0, 0, 255, 255, 255);
            ST7735R_DrawText(START_X, START_Y + 3 * STEP_Y, "HARD", 0, 0, 0, 255, 255, 255);
        }
        else if(lvl == 1)
        {
            ST7735R_DrawText(START_X, START_Y + 1 * STEP_Y, "EASY", 0, 0, 0, 255, 255, 255);
            ST7735R_DrawText(START_X, START_Y + 2 * STEP_Y, "MEDIUM", 0, 0, 0, 255, 0, 0);
            ST7735R_DrawText(START_X, START_Y + 3 * STEP_Y, "HARD", 0, 0, 0, 255, 255, 255);
        }
        else if(lvl == 2)
        {
            ST7735R_DrawText(START_X, START_Y + 1 * STEP_Y, "EASY", 0, 0, 0, 255, 255, 255);
            ST7735R_DrawText(START_X, START_Y + 2 * STEP_Y, "MEDIUM", 0, 0, 0, 255, 255, 255);
            ST7735R_DrawText(START_X, START_Y + 3 * STEP_Y, "HARD", 0, 0, 0, 255, 0, 0);
        }
        _delay_ms(150);
    }

    if(lvl == 0)
    {
        width = 11;
        height = 6;
        squares_left = 66;
        bombs = 13;
        start_x = START_X + STEP_X;
    }
    else if(lvl == 1)
    {
        width = 13;
        height = 7;
        squares_left = 91;
        bombs = 20;
        start_x = START_X;
    }
    else if(lvl == 2)
    {
        width = 15;
        height = 8;
        squares_left = 120;
        bombs = 30;
        start_x = START_X - STEP_X;
    }

    compute_board();
    // Activate interupts
    ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 255, 255, 255);
    draw_board();
    sei();
    game_on = 1;
}

ISR(PCINT0_vect)
{
    uint8_t i, j;

    if(!game_on)
        return;

    if ((PINA & (1 << PA3)) != 0)
    {
        pointer_x = (pointer_x + 1) % width;
    }
    else if ((PINA & (1 << PA7)) != 0)
    {
        if(pointer_y == 0)
            pointer_y = height;

        pointer_y -= 1;
    }
    else if ((PINA & (1 << PA5)) != 0)
    {
        if(pointer_x == 0)
            pointer_x = width;

        pointer_x -= 1;
    }
    else if ((PINA & (1 << PA4)) != 0)
    {
        pointer_y = (pointer_y + 1) % height;
    }
    else if ((PINA & (1 << PA6)) != 0)
    {
        if(marks[pointer_y][pointer_x] == 0) {
            if(board[pointer_y][pointer_x] == 0)
                fill((int8_t)pointer_y, (int8_t)pointer_x);
            else{
                marks[pointer_y][pointer_x] = 1;
                squares_left --;
            }

            if(board[pointer_y][pointer_x] == -1){
                for(i = 0; i < height; i++)
                    for(j = 0; j < width; j++)
                        marks[i][j] = 1;
                
                losses += 1;
                draw_board();
                _delay_ms(5000);
                game_init();
                return;     
            }

            if(squares_left == bombs){
                for(i = 0; i < height; i++)
                    for(j = 0; j < width; j++)
                        marks[i][j] = 1;
                
                wins += 1;
                draw_board();
                _delay_ms(5000);
                game_init();
                return;
            }
        }
    }

    draw_board();
    _delay_ms(150);
}

ISR(PCINT1_vect)
{
    if(!game_on)
        return;

    if ((PINB & (1 << PB2)) == 0)
    {
        if(marks[pointer_y][pointer_x] == 2)
            marks[pointer_y][pointer_x] = 0;
        else if(marks[pointer_y][pointer_x] == 0)
            marks[pointer_y][pointer_x] = 2;
        
        _delay_ms(150);
    }

    draw_board();
}

void IO_init()
{
    // Init communication channel
    SPI_init();

    // Init LCD
    ST7735R_Begin();

    // Set buttons and pull-ups resistence
    DDRA  &= ~(1 << PA3);
    PORTA |= (1 << PA3);

    DDRA  &= ~(1 << PA4);
    PORTA |= (1 << PA4);

    DDRA  &= ~(1 << PA5);
    PORTA |= (1 << PA5);

    DDRA  &= ~(1 << PA6);
    PORTA |= (1 << PA6);

    DDRA  &= ~(1 << PA7);
    PORTA |= (1 << PA7);

    DDRB &= ~(1 << PB2);
    PORTB |= (1 << PB2);

    // Activate PIN CHANGE INTERRUPT ENABLE for PCINT7:0 (port A)
    PCICR |= (1 << PCIE0) | (1 << PCIE1);
    // Activate interupts for buttons PA3 -> PA7
    PCMSK0 |= (1 << PCINT3) | (1 << PCINT4) | (1 << PCINT5) | (1 << PCINT6) | (1 << PCINT7);
    PCMSK1 |= (1 << PCINT10);

    // Clear screen with WHITE.
    ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 255, 255, 255);
}


int main(void)
{
    srand(0);
    IO_init();
    game_init();

    while(1){}

    return 0;
}
